// Copyright (C) 2007-2009 David Kment
// For license see license.txt

#include "CGTASceneNode.h"
#include "main.h"

#include "ISceneManager.h"
#include "IVideoDriver.h"
#include "IFileSystem.h"
#include "IReadFile.h"
#include "S3DVertex.h"

CGTASceneNode::CGTASceneNode(IrrlichtDevice *device, int gameId, ISceneNode* parent, ISceneManager* mgr,
		s32 id, const core::vector3df& position,
		const core::vector3df& rotation, const core::vector3df& scale)
	: ISceneNode(parent, mgr, id, position, rotation, scale)
{
    GameId = gameId;

    Device = device;

	Logger = Device->getLogger();

    IdeMode = 0;
    IplMode = 0;

    int maxObjectIds;

    if(GameId == GAMEID_GTA3) {maxObjectIds = MAXOBJECTIDS_GTA3;}
    if(GameId == GAMEID_GTAVC) {maxObjectIds = MAXOBJECTIDS_GTAVC;}
    if(GameId == GAMEID_GTASA) {maxObjectIds = MAXOBJECTIDS_GTASA;}

    Device->getSceneManager()->addEmptySceneNode();

    int i;
    ObjectDefinition *temp;
    for(i=0;i<maxObjectIds;i++)
    {
        temp = new ObjectDefinition(Device->getSceneManager());
        Definitions.push_back(*temp);
    }
}

CGTASceneNode::~CGTASceneNode()
{
	if (Logger) {Logger->drop();}
}

void CGTASceneNode::render()
{
}

const core::aabbox3d<f32>& CGTASceneNode::getBoundingBox() const
{
    return Box;
}

void CGTASceneNode::OnRegisterSceneNode()
{
	if (IsVisible)
	{
		SceneManager->registerNodeForRendering(this);
		ISceneNode::OnRegisterSceneNode();
	}
}

void CGTASceneNode::initialise()
{
    // every GTA from series 3 has a default.dat
    loadDAT("data/default.dat");

    stringc startFile = "";

    if(GameId == GAMEID_GTA3) {startFile = "data/gta3.dat";}
    if(GameId == GAMEID_GTAVC) {startFile = "data/gta_vc.dat";}
    if(GameId == GAMEID_GTASA) {startFile = "data/gta.dat";}

    loadDAT(startFile);

    // fixme: this is a pretty fragile way of getting the GTA root node
    ISceneNode *root = *SceneManager->getRootSceneNode()->getChildren().begin();

    root->setRotation(core::vector3df(-90.0f, 0.0f, 0.0f));
    // root->setPosition(core::vector3df(0.0f, -150.0f, 0.0f));
}

std::vector<CGTASceneNode::ObjectInstance> *CGTASceneNode::getObjectInstances()
{
    return &Instances;
}

void CGTASceneNode::loadDAT(stringc inFileName) {processFile(inFileName, FORMAT_DAT);}

void CGTASceneNode::loadIDE(stringc inFileName) {processFile(inFileName, FORMAT_IDE);}

void CGTASceneNode::loadIPL(stringc inFileName) {processFile(inFileName, FORMAT_IPL);}

void CGTASceneNode::processFile(stringc inFileName, int mode)
{
    inFileName.replace('\\', '/');

    Logger->log("Loading file", inFileName.c_str(), ELL_INFORMATION);

    io::IReadFile* file = Device->getFileSystem()->createAndOpenFile(inFileName.c_str());

    if(!file)
    {
        Logger->log("Cannot open", inFileName.c_str(), ELL_WARNING);
        return;
    }

    // here we read bytewise. this should be optimized.
    // if only irrlicht had a readLine function...

    stringc line = "";
    c8 tmp[2];
    tmp[1] = 0;
    int bytesRead = 0;

    do
    {
        bytesRead = file->read(tmp, 1);

        if(bytesRead > 0)
        {
            if(tmp[0] == '\r') {}
            else if(tmp[0] == '\n')
            {
                // line end
                if(!startsWith(&line, "#") && line.size() > 0)
                {
                    if(mode == FORMAT_DAT) {processDAT(line);}
                    if(mode == FORMAT_IPL) {processIPL(line);}
                    if(mode == FORMAT_IDE) {processIDE(line);}
                }

                line = "";
            }
            else
            {
                // append character to string
                line += tmp;
            }
        }
    } while(bytesRead > 0);

    file->drop();
}

void CGTASceneNode::processDAT(stringc inStr)
{
    if(startsWith(&inStr, "SPLASH ")) {/* SPLASH command is currently ignored */}
    else if(startsWith(&inStr, "COLFILE ")) {/* COLFILE command is currently ignored */}
    else if(startsWith(&inStr, "MAPZONE ")) {/* MAPZONE command is currently ignored */}
    else if(startsWith(&inStr, "TEXDICTION "))
    {
        Device->getFileSystem()->registerFileArchive(inStr.subString(11, inStr.size() - 11), true, false); // " 11 = "TEXDICTION "
    }
    else if(startsWith(&inStr, "IDE "))
    {
        loadIDE(inStr.subString(4, inStr.size() - 4)); // " 4 = "IDE "
    }
    else if(startsWith(&inStr, "IPL "))
    {
        loadIPL(inStr.subString(4, inStr.size() - 4)); // " 4 = "IPL "
    }
    else
    {
        Logger->log("Unknown command", inStr.c_str(), ELL_WARNING);
    }
}

void CGTASceneNode::processIPL(stringc inStr)
{
    stringc iplTempName;
    int iplTempId = 0, ipl_temp_int = 0, ipl_temp_lod = 0;
    double iplTempPosX = 0.0f, iplTempPosY = 0.0f, iplTempPosZ = 0.0f;
    double iplTempScaleX = 0.0f, iplTempScaleY = 0.0f, iplTempScaleZ = 0.0f;
    double iplTempRotX = 0.0f, iplTempRotY = 0.0f, iplTempRotZ = 0.0f, ipl_temp_rot_w = 0.0f;

    double q2aOutX = 0.0f, q2aOutY = 0.0f, q2aOutZ = 0.0f;

    if(startsWith(&inStr, "inst")) {IplMode = IPL_MODE_INST;}
    else if(startsWith(&inStr, "cull")) {IplMode = IPL_MODE_CULL;}
    else if(startsWith(&inStr, "pick")) {IplMode = IPL_MODE_PICK;}
    else if(startsWith(&inStr, "end")) {IplMode = 0;}
    else
    {
        if(IplMode == IPL_MODE_INST)
        {
            int i = 0;
            char *tmp = (char *)inStr.c_str();
            char *tok = strtok(tmp, " ,");

            while(tok)
            {
                if(GameId == GAMEID_GTA3)
                {
                    if(i==0) {iplTempId = atoi(tok);}
                    if(i==1) {iplTempName = tok;}
                    if(i==2) {iplTempPosX = atof(tok);}
                    if(i==3) {iplTempPosY = atof(tok);}
                    if(i==4) {iplTempPosZ = atof(tok);}
                    if(i==5) {iplTempScaleX = atof(tok);}
                    if(i==6) {iplTempScaleY = atof(tok);}
                    if(i==7) {iplTempScaleZ = atof(tok);}
                    if(i==8) {iplTempRotX = atof(tok);}
                    if(i==9) {iplTempRotY = atof(tok);}
                    if(i==10) {iplTempRotZ = atof(tok);}
                    if(i==11) {ipl_temp_rot_w = atof(tok);}
                }
                else if(GameId == GAMEID_GTAVC)
                {
                    if(i==0) {iplTempId = atoi(tok);}
                    if(i==1) {iplTempName = tok;}
                    if(i==2) {ipl_temp_int = atoi(tok);}
                    if(i==3) {iplTempPosX = atof(tok);}
                    if(i==4) {iplTempPosY = atof(tok);}
                    if(i==5) {iplTempPosZ = atof(tok);}
                    if(i==6) {iplTempScaleX = atof(tok);}
                    if(i==7) {iplTempScaleY = atof(tok);}
                    if(i==8) {iplTempScaleZ = atof(tok);}
                    if(i==9) {iplTempRotX = atof(tok);}
                    if(i==10) {iplTempRotY = atof(tok);}
                    if(i==11) {iplTempRotZ = atof(tok);}
                    if(i==12) {ipl_temp_rot_w = atof(tok);}
                }
                else if(GameId == GAMEID_GTASA)
                {
                    if(i==0) {iplTempId = atoi(tok);}
                    if(i==1) {iplTempName = tok;}
                    if(i==2) {ipl_temp_int = atoi(tok);}
                    if(i==3) {iplTempPosX = atof(tok);}
                    if(i==4) {iplTempPosY = atof(tok);}
                    if(i==5) {iplTempPosZ = atof(tok);}
                    if(i==6) {iplTempRotX = atof(tok);}
                    if(i==7) {iplTempRotY = atof(tok);}
                    if(i==8) {iplTempRotZ = atof(tok);}
                    if(i==9) {ipl_temp_rot_w = atof(tok);}
                    if(i==10) {ipl_temp_rot_w = atof(tok);}
                    if(i==11) {ipl_temp_lod = atoi(tok);}
                }

                i++;

                tok = strtok(NULL, " ,");
            }

            quaternion *quat = new quaternion(iplTempRotX, iplTempRotY, iplTempRotZ, ipl_temp_rot_w);
            vector3df *euler = new vector3df();
            quat->toEuler(*euler);
            q2aOutX = euler->X;
            q2aOutY = euler->Y;
            q2aOutZ = euler->Z;
            delete quat;
            delete euler;

            ObjectInstance *instance = new ObjectInstance(Device->getSceneManager());

            instance->setName(iplTempName);
            instance->setTextureDir(Definitions[iplTempId].getTextureDir());
            instance->setDrawDistance(Definitions[iplTempId].getDrawDistance());
            instance->setID(iplTempId);

            instance->setPosition(core::vector3df(iplTempPosX, iplTempPosY, iplTempPosZ));
            instance->setRotation(core::vector3df(q2aOutX, q2aOutY, q2aOutZ));
            instance->setScale(core::vector3df(iplTempScaleX, iplTempScaleY, iplTempScaleZ));

            if(!DYNAMIC_LOD_MODE)
            {
                instance->update();
            }

            Instances.push_back(*instance);
        }
    }
}

void CGTASceneNode::processIDE(stringc inStr)
{
    int ideTempId = 0;
    stringc ideTempName;
    stringc ideTempTextureDir;
    int ideTempObjectCount;
    float ideTempDrawDistance;

    if(startsWith(&inStr, "objs")) {IdeMode = IDE_MODE_OBJS;}
    else if(startsWith(&inStr, "tobj")) {IdeMode = IDE_MODE_TOBJ;}
    else if(startsWith(&inStr, "path")) {IdeMode = IDE_MODE_PATH;}
    else if(startsWith(&inStr, "2dfx")) {IdeMode = IDE_MODE_2DFX;}
    else if(startsWith(&inStr, "end")) {IdeMode = 0;}
    else
    {
        if(IdeMode == IDE_MODE_OBJS)
        {
            int i = 0;
            char *tmp = (char *)inStr.c_str();
            char *tok = strtok(tmp, " ,");

            while(tok)
            {
                if(i==0) {ideTempId = atoi(tok);}
                if(i==1) {ideTempName = tok;}
                if(i==2) {ideTempTextureDir = tok;}
                if(i==3) {ideTempObjectCount = atoi(tok);}
                if(i==4) {ideTempDrawDistance = atof(tok);}

                i++;

                tok = strtok(NULL, " ,");
            }

            Definitions[ideTempId].setName(ideTempName);
            Definitions[ideTempId].setTextureDir(ideTempTextureDir);
            Definitions[ideTempId].setDrawDistance(ideTempDrawDistance);
        }
    }
}

CGTASceneNode::ObjectInstance::ObjectInstance(ISceneManager *sceneManager)
{
    SceneManager = sceneManager;

    Id = -1;

    Mesh = NULL;
    Node = NULL;

    Name = "";

    Position = core::vector3df(0.0f, 0.0f, 0.0f);
    Rotation = core::vector3df(0.0f, 0.0f, 0.0f);
    Scale = core::vector3df(0.0f, 0.0f, 0.0f);

    Render = true;
}

CGTASceneNode::ObjectInstance::~ObjectInstance()
{
}

void CGTASceneNode::ObjectInstance::update()
{
    if(Render)
    {
        if(Node == NULL)
        {
            if(Mesh == NULL)
            {
                loadFromDFFFile();
            }

            if(Mesh)
            {
                // fixme: this is a pretty fragile way of getting the GTA root node
                ISceneNode *root = *SceneManager->getRootSceneNode()->getChildren().begin();

                Node = SceneManager->addMeshSceneNode(Mesh, root);

                Node->setPosition(Position);
                Node->setRotation(Rotation);
                // Node->setScale(Scale);
            }
        }
    }
    else
    {
        if(Node)
        {
            Node->remove();
            Node = NULL;
        }

        if(Mesh)
        {
            Mesh = NULL;
        }
    }
}

void CGTASceneNode::ObjectInstance::loadFromDFFFile()
{
    stringc completeFileName = "";
    completeFileName += SceneManager->getParameters()->getAttributeAsString("GTA_MODELS_PATH");
    completeFileName += Name;
    completeFileName += ".dff";

    IsLODMesh = false;

    // fixme: this LOD detection works for GTA3/VC only

    stringc temp = Name.subString(0, 3);

    if(temp.equals_ignore_case("lod"))
    {
        IsLODMesh = true;
    }

    if(!DYNAMIC_LOD_MODE)
    {
        if((LOWDETAIL_MESHES_ONLY && !IsLODMesh) || HIDETAIL_MESHES_ONLY && IsLODMesh)
        {
            Render = false;
            return;
        }
    }

    stringc texturePath = "";
    texturePath += SceneManager->getParameters()->getAttributeAsString("GTA_TEXTURES_PATH");
    texturePath += TextureDir;
    texturePath += "/";

    SceneManager->getParameters()->setAttribute("DFF_TEXTURE_PATH", texturePath.c_str());

    // fixme: force load of TXD file
    SceneManager->getFileSystem()->registerFileArchive(TextureDir + ".txd", true, false);

    Mesh = SceneManager->getMesh(completeFileName.c_str());

    if(!Mesh)
    {
        // error loading mesh
    }
}

float CGTASceneNode::ObjectInstance::getDrawDistance()
{
    return DrawDistance;
}

int CGTASceneNode::ObjectInstance::getID()
{
    return Id;
}

core::vector3df CGTASceneNode::ObjectInstance::getPosition()
{
    return Position;
}

bool CGTASceneNode::ObjectInstance::getRender()
{
    return Render;
}

bool CGTASceneNode::ObjectInstance::isLODMesh()
{
    return IsLODMesh;
}

void CGTASceneNode::ObjectInstance::setDrawDistance(float distance)
{
    DrawDistance = distance;
}

void CGTASceneNode::ObjectInstance::setName(stringc name)
{
    Name = name;
}

void CGTASceneNode::ObjectInstance::setTextureDir(stringc textureDir)
{
    TextureDir = textureDir;
}

void CGTASceneNode::ObjectInstance::setID(int id)
{
    Id = id;
}

void CGTASceneNode::ObjectInstance::setPosition(core::vector3df position)
{
    Position = position;
}

void CGTASceneNode::ObjectInstance::setRotation(core::vector3df rotation)
{
    Rotation = rotation;
}

void CGTASceneNode::ObjectInstance::setScale(core::vector3df scale)
{
    Scale = scale;
}

void CGTASceneNode::ObjectInstance::setRender(bool render)
{
    Render = render;
    update();
}

CGTASceneNode::ObjectDefinition::ObjectDefinition(ISceneManager *sceneManager)
{
    SceneManager = sceneManager;

    Name = "";
    TextureDir = "";
}

CGTASceneNode::ObjectDefinition::~ObjectDefinition()
{

}

float CGTASceneNode::ObjectDefinition::getDrawDistance()
{
    return DrawDistance;
}

stringc CGTASceneNode::ObjectDefinition::getTextureDir()
{
    return TextureDir;
}

void CGTASceneNode::ObjectDefinition::setDrawDistance(float distance)
{
    DrawDistance = distance;
}

void CGTASceneNode::ObjectDefinition::setName(stringc name)
{
    Name = name;
}

void CGTASceneNode::ObjectDefinition::setTextureDir(stringc textureDir)
{
    TextureDir = textureDir;
}

bool startsWith(stringc *haystack, c8 *needle)
{
    return haystack->trim().equalsn(needle, strlen(needle));
}
