// Copyright (C) 2007-2009 David Kment
// For license see license.txt

#include "CIMGArchiveLoader.h"

#include "IReadFile.h"

CIMGArchiveLoader::CIMGArchiveLoader(io::IFileSystem* fs)
{
    FileSystem = fs;
}

CIMGArchiveLoader::~CIMGArchiveLoader()
{
}

bool CIMGArchiveLoader::isALoadableFileFormat(const core::string<c16>& filename) const
{
	return core::hasFileExtension(filename, "img");
}

bool CIMGArchiveLoader::isALoadableFileFormat(io::IReadFile* file) const
{
    // since all IMG files must have extension .img, we can skip this part and return false
    return false;
}

IFileArchive* CIMGArchiveLoader::createArchive(const core::string<c16>& filename, bool ignoreCase, bool ignorePaths) const
{
	IFileArchive *archive = 0;
	io::IReadFile* file = FileSystem->createAndOpenFile(filename);

	if (file)
	{
		archive = createArchive(file, ignoreCase, ignorePaths);
		file->drop();
	}
	
	return archive;
}

IFileArchive* CIMGArchiveLoader::createArchive(io::IReadFile* file, bool ignoreCase, bool ignorePaths) const
{
	IFileArchive *archive = 0;
	if (file)
	{
		file->seek(0);
		archive = new CIMGReader(file, ignoreCase, ignorePaths);
        ((CIMGReader *)archive)->setFileSystem(FileSystem);
        ((CIMGReader *)archive)->init();
	}
	return archive;
}

CIMGReader::CIMGReader(IReadFile* file, bool ignoreCase, bool ignorePaths)
{
    Type = "img";
    File = file;
    IgnoreCase = ignoreCase;
    IgnorePaths = ignorePaths;

	if(File)
	{
		File->grab();
	}
}

CIMGReader::~CIMGReader()
{
	if (File)
    {
        File->drop();
    }
}

void CIMGReader::init()
{
    SIMGFileHeaderV2 header;

    File->read(&header, sizeof(SIMGFileHeaderV2));

    if(header.fourcc[0] == 'V' && header.fourcc[1] == 'E' && header.fourcc[2] == 'R' && header.fourcc[3] == '2')
    {
        // v2 IMG file

        // load directory into memory

        SIMGDirEntry dirEntry;
        SIMGFileEntry fileEntry;

        u32 i;
        for(i=0; i < header.entries; i++)
        {
            File->read(&dirEntry, sizeof(SIMGDirEntry));

            fileEntry.simpleFileName = dirEntry.name;
            fileEntry.pos = dirEntry.pos * 2048;
            fileEntry.length = dirEntry.length * 2048;

            if(IgnoreCase)
            {
                fileEntry.simpleFileName.make_lower();
            }

            FileList.push_back(fileEntry);
        }
    }
    else
    {
        IReadFile* DirFile;

        string<c16> dirFileName = File->getFileName();
        s32 offset = dirFileName.findLast('.');
        dirFileName[offset+1] = 'd';
        dirFileName[offset+2] = 'i';
        dirFileName[offset+3] = 'r';

        DirFile = FileSystem->createAndOpenFile(dirFileName);

        if (DirFile)
        {
            // v1 IMG file

            // load directory into memory

            SIMGDirEntry dirEntry;
            SIMGFileEntry fileEntry;

            while(DirFile->read(&dirEntry, sizeof(SIMGDirEntry)) == sizeof(SIMGDirEntry))
            {
                fileEntry.simpleFileName = dirEntry.name;
                fileEntry.pos = dirEntry.pos * 2048;
                fileEntry.length = dirEntry.length * 2048;

                if(IgnoreCase)
                {
                    fileEntry.simpleFileName.make_lower();
                }

                FileList.push_back(fileEntry);
            }
        }
        else
        {
            // unsupported IMG file, or DIR file not found
        }
    }
}

IReadFile* CIMGReader::openFile(const core::string<c16>& filename)
{
	s32 index = findFile(filename);

	if (index != -1)
    {
        return openFile(index);
    }

	return 0;
}

IReadFile* CIMGReader::openFile(s32 index)
{
    return FileSystem->createLimitReadFile(FileList[index].simpleFileName, File, FileList[index].pos, FileList[index].length);
}

u32 CIMGReader::getFileCount() const
{
	return FileList.size();
}

const IFileArchiveEntry* CIMGReader::getFileInfo(u32 index)
{
	return &FileList[index];
}

s32 CIMGReader::findFile(const core::string<c16>& filename)
{
    string<c16> simpleFileName = filename;

	if (IgnoreCase)
		simpleFileName.make_lower();

	if (IgnorePaths)
		core::deletePathFromFilename(simpleFileName);

    s32 res = -1;
    u32 i;

    // this way of comparison is slow, but binary_search was buggy for me
    for(i=0; i < FileList.size(); i++)
    {
        if(FileList[i].simpleFileName == simpleFileName)
        {
            res = i;
            break;
        }
    }

	return res;
}

void CIMGReader::setFileSystem(io::IFileSystem* fs)
{
    FileSystem = fs;
}
