// Copyright (C) 2007-2009 David Kment
// For license see license.txt

#include "CTXDArchiveLoader.h"

#include "IReadFile.h"

CTXDArchiveLoader::CTXDArchiveLoader(io::IFileSystem* fs)
{
    FileSystem = fs;
}

CTXDArchiveLoader::~CTXDArchiveLoader()
{
}

bool CTXDArchiveLoader::isALoadableFileFormat(const core::string<c16>& filename) const
{
	return core::hasFileExtension(filename, "TXD");
}

bool CTXDArchiveLoader::isALoadableFileFormat(io::IReadFile* file) const
{
    // since all TXD files must have extension .TXD, we can skip this part and return false
    return false;
}

IFileArchive* CTXDArchiveLoader::createArchive(const core::string<c16>& filename, bool ignoreCase, bool ignorePaths) const
{
	IFileArchive *archive = 0;
	io::IReadFile* file = FileSystem->createAndOpenFile(filename);

	if (file)
	{
		archive = createArchive(file, ignoreCase, ignorePaths);
		file->drop();
	}
	
	return archive;
}

IFileArchive* CTXDArchiveLoader::createArchive(io::IReadFile* file, bool ignoreCase, bool ignorePaths) const
{
	IFileArchive *archive = 0;

    if (file)
    {
        file->seek(0);
        archive = new CTXDReader(file, ignoreCase, ignorePaths);
        ((CTXDReader *)archive)->setFileSystem(FileSystem);
    }

	return archive;
}

CTXDReader::CTXDReader(IReadFile* file, bool ignoreCase, bool ignorePaths)
{
    Type = "TXD";
    File = file;
    IgnoreCase = ignoreCase;
    IgnorePaths = ignorePaths;
    TXDName = file->getFileName();
    core::deletePathFromFilename(TXDName);
    TXDName = TXDName.subString(0, TXDName.size() - 4);

	if(File)
	{
		File->grab();

        parseRWSection(File->getSize(), 0, 0);
	}
}

CTXDReader::~CTXDReader()
{
	if (File)
    {
        File->drop();
    }
}

IReadFile* CTXDReader::openFile(const core::string<c16>& filename)
{
	s32 index = findFile(filename);

	if (index != -1)
    {
        return openFile(index);
    }

	return 0;
}

IReadFile* CTXDReader::openFile(s32 index)
{
    return FileSystem->createLimitReadFile(FileList[index].simpleFileName, File, FileList[index].pos, FileList[index].length);
}

u32 CTXDReader::getFileCount() const
{
	return FileList.size();
}

const IFileArchiveEntry* CTXDReader::getFileInfo(u32 index)
{
	return &FileList[index];
}

s32 CTXDReader::findFile(const core::string<c16>& filename)
{
    string<c16> simpleFileName = filename;

	if (IgnoreCase)
		simpleFileName.make_lower();

	if (IgnorePaths)
		core::deletePathFromFilename(simpleFileName);

    s32 res = -1;
    u32 i;

    // this way of comparison is slow, but binary_search was buggy for me
    for(i=0; i < FileList.size(); i++)
    {
        if(FileList[i].simpleFileName == simpleFileName)
        {
            res = i;
            break;
        }
    }

	return res;
}

void CTXDReader::parseRWSection(int size, int depthlevel, int parentType)
{
    if(size == 0) {return;}

    unsigned int sectionType, sectionSize;
    unsigned short sectionunknown = 0, sectionVersion = 0;

    int endoffset = File->getPos() + size;

    while(File->getPos() < endoffset)
    {
        File->read(&sectionType, sizeof(int));
        File->read(&sectionSize, sizeof(int));
        File->read(&sectionunknown, sizeof(short));
        File->read(&sectionVersion, sizeof(short));

        switch(sectionType)
        {
            case DFF_RW_DATA:
                processRWData(sectionSize, parentType, sectionVersion);
                break;

            // all below can contain more sections
            case DFF_RW_EXTENSION:
                parseRWSection(sectionSize, depthlevel + 1, sectionType);
                break;

            case DFF_RW_DICTIONARY:
                parseRWSection(sectionSize, depthlevel + 1, sectionType);
                break;

            case DFF_RW_TEXTURE_NATIVE:
                parseRWSection(sectionSize, depthlevel + 1, sectionType);
                break;

            default:
                // unknown section type, skip sectionSize bytes
                File->seek(sectionSize, true);
                break;
        }
    }
}

void CTXDReader::processRWData(int size, int type, int version)
{
    if(type == DFF_RW_TEXTURE_NATIVE)
    {
        long pos = File->getPos();

        STXDTextureHeader header;

        File->read(&header, sizeof(STXDTextureHeader));

        File->seek(size - sizeof(STXDTextureHeader), true);

        STXDFileEntry fileEntry;

        fileEntry.simpleFileName = TXDName;
        fileEntry.simpleFileName += "/";
        fileEntry.simpleFileName += header.diffuseTextureName;
        fileEntry.simpleFileName += ".gtatexture";
        fileEntry.pos = pos;
        fileEntry.length = size;

        if(IgnoreCase)
        {
            fileEntry.simpleFileName.make_lower();
        }

        FileList.push_back(fileEntry);

        // repeat for alphaTextureName

        fileEntry.simpleFileName = TXDName;
        fileEntry.simpleFileName += "/";
        fileEntry.simpleFileName += header.alphaTextureName;
        fileEntry.simpleFileName += ".gtatexture";

        if(IgnoreCase)
        {
            fileEntry.simpleFileName.make_lower();
        }

        FileList.push_back(fileEntry);
    }
    else
    {
        // unknown data section type, skip size bytes
        File->seek(size, true);
    }
}

void CTXDReader::setFileSystem(io::IFileSystem* fs)
{
    FileSystem = fs;
}
